/* dm-steg.h - definitions for dm-steg kernel module.
 * all u64's in on-disk and dual-purpose structures are little endian.
 * memory-only u64's are host byte order */

#define TO_ARRAY		0
#define FROM_ARRAY		1

#define MINIMUM_BLOCK_BITS	16
#define MINIMUM_BLOCK_BYTES	(1 << MINIMUM_BLOCK_BITS)
#define MINIMUM_ATOM_BYTES	512 

#define KEY_BYTES		32
#define SALT_BYTES		56
#define BUNNY_ENTRY_BYTES	64
#define META_HASH_BYTES		24
#define HEADER_HASH_BYTES	32
#define ASPECT_NAME_BYTES	64

#define KEYFRAG_SHIFT		5
#define KEYFRAG_BYTES		(1 << KEYFRAG_SHIFT)
#define KEYFRAG_PAGE_SHIFT	(PAGE_SHIFT - KEYFRAG_SHIFT)
#define KEYFRAG_PAGE_MASK	(~((1 << KEYFRAG_PAGE_SHIFT) - 1))

#define BLOCK_SHIFT		KEYFRAG_SHIFT
#define BLOCK_PAGE_SHIFT	KEYFRAG_PAGE_SHIFT
#define BLOCK_PAGE_MASK		KEYFRAG_PAGE_MASK

#define U64_PAGE_SHIFT		(PAGE_SHIFT - 3)
#define U64_PAGE_MASK		(~((1 << U64_PAGE_SHIFT) - 1))

#define ASPECT_HEADER_VERSION	1
#define ASPECT_HEADER_MAGIC	0x21616172646e7550

#define MESG_STR(x)		x, sizeof(x)

struct keyfrag {
	u8 d[KEYFRAG_BYTES];
};

/* offset is byte offset into block device and also first 8 bytes of block
 * keyfrag. Little endian. */
struct block {
	u64 offset;
	u8 rest[KEYFRAG_BYTES - sizeof(u64)];
};

/* pyramid format: upper 2 bits: 01 = data block, 10 = index block,
 * 11 = the header block. lower 62 bits = block num within index */
#define MAX_LAYERS		(65 - MINIMUM_BLOCK_BITS)
#define pyramid_type(x)		((x)&((u64)3<<62))
#define pyramid_blocknum(x)	((x)&~((u64)3<<62))
#define PYRAMID_HEADER		((u64)2<<62)
#define PYRAMID_DATA		((u64)1<<62)
#define PYRAMID_FREE_BLOCK	0

#define STEG_UNLOCKED		(-1)
#define STEG_ALL_LOCKED		(-2)

struct bio_meta {
	wait_queue_head_t *q;
	atomic_t incomplete;
	int error;
};

struct steg_request {
	struct work_struct work;	/* for encrypt/decrypt workqueues */
	struct list_head list;		/* for adding to a->fifo */
	struct aspect *a;		/* aspect to which request belongs */
	struct bio *bio;		/* given to steg_map by DM layer */
	struct bio *original;		/* for encrypted writes */
	atomic_t *requests_running;	/* we decrement this when finished */
	atomic_t incomplete;		/* number of bios generated */
	int error;
};

/* journal logs shuffles to avoid data loss or metadata corruption if power
 * failure occurs. Also keeps track of alignment pyramid so blocks do not
 * become less aligned. All little endian */
struct aspect_disk_journal {
	u64 block_being_moved;		/* or 0 if none */
	u64 src_offset;			/* offset into block device */
	u64 dst_offset;			/* offset into block device */
	u64 ascending;			/* if 1, next layer is up (layer--) */
	u64 shuffles_left;		/* shuffles left before next layer */
	u64 promoted[0];		/* 0-term'd list of promoted blocks */
};

/* All host byte order */
struct aspect {
	struct dm_target *ti;
	struct dm_dev *device;

	struct list_head *fifo;
	struct workqueue_struct *decrypt_queue;
	struct workqueue_struct *encrypt_queue;
	struct workqueue_struct *shuffle_queue;

	struct block **block;
	struct keyfrag **atom_keyfrag;

	struct block **header_block;
	struct block *header_block_page_table;
	struct block header_block_data;

	/* From header */
	unsigned int version;
	unsigned int sequence;
	unsigned int blocks;
	unsigned int block_bytes;
	unsigned int atom_bytes;
	unsigned int encryption;
	unsigned int shuffling;
	unsigned int journalling;
	unsigned int journal_offset;
	unsigned int keyfrags_offset;
	unsigned int index_offset;
	unsigned int parent_level;
	u8 parent_pass_hash[KEY_BYTES];
	u8 salt[SALT_BYTES];
	char name[ASPECT_NAME_BYTES];

	/* Calculated */
	u64 bytes;				/* Aspect internal size */
	u64 block_mask;
	u64 atom_mask;
	unsigned int block_shift;
	unsigned int atom_shift;
	unsigned int meta_atom_bytes;
	unsigned int atoms_per_block;

	/* For shuffling */
	struct aspect_disk_journal *journal;
	u64 ***pyramid;
	u64 *free_pyramid_block;
	u64 *shuffle_target;
	unsigned int *blocks_in_layer;
	void *random_page;
	u8 bunny_seed[BUNNY_ENTRY_BYTES];
	u8 bunny_output[BUNNY_ENTRY_BYTES];
	u8 header_key[KEY_BYTES];		/* Set once, on aspect load */
	u8 header_rndhash[64];			/* Set once, on aspect load */
	unsigned int layers;
	unsigned int free_block_layer;
	unsigned int bunny_pair_ready;
	unsigned int journal_open;
	unsigned int blocks_promoted;

	/* For locking */
	atomic_t open_interest;
	spinlock_t rw_metalock;
	atomic_t requests_running_array[2];
	atomic_t *requests_running;
	atomic_t *old_requests_running;
	wait_queue_head_t requests_running_q;
	u64 read_locked_offset;
	u64 write_locked_offset;
	atomic_t writes_in_flight;
	struct rw_semaphore slowpath_lock;
	struct semaphore shuffle_lock;
};

/* Encrypted and stored in header sector */
/* All u64s little endian */
struct aspect_disk_header {
	u64 version;				/* Encryption starts here */
	u64 sequence;				/* For header block shuffles */
	u64 blocks;				/* Number of data blocks */
	u64 block_bytes;			/* Size of blocks */
	u64 atom_bytes;				/* Size of constituent atoms */
	u64 encryption;				/* For data blocks only */
	u64 shuffling;				/* Boolean */
	u64 journalling;			/* Boolean */
	u64 journal_offset;			/* In bytes */
	u64 keyfrags_offset;			/* ditto */
	u64 index_offset;			/* ditto */
	u64 parent_level;			/* or -1 for no parent */

	u8 parent_pass_hash[KEY_BYTES];		/* for loading parent aspect */

	struct block header_block_data;		/* for bootstrapping */
	struct keyfrag seed_keyfrag;		/* for bootstrapping */
	u8 salt[SALT_BYTES];			/* for hash input padding */

	char name[ASPECT_NAME_BYTES];
	char padding[104];
	u8 inner_hash[32];			/* SHA-256 of plaintext */
};

/* Header sector format. 512 bytes total */
struct aspect_disk_header_enc {
	u8 bunny_seed[BUNNY_ENTRY_BYTES];
	struct aspect_disk_header header;
};

/* At end of every meta atom */
struct meta_tail {
	u8 hash[META_HASH_BYTES];
	u64 rnd64;
};
